////////////////////////////////////////////////////////////////////////////
//
// COPYRIGHT ANSYS. ALL RIGHTS RESERVED.
//
// $LastChangedDate$
// $LastChangedRevision$
// $LastChangedBy$
//
// Author: Lionel TOMASO
//
// Decription : implementation of ISamplingnMethod and the CSharpSampling
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net.Sockets;
using Ans.ComponentSystem;
using Ansys.DesignXplorer.API.Common;
using Ansys.DesignXplorer.API.Optimization;
using Ansys.DesignXplorer.API.Sampling;

namespace Ansys.DesignXplorer.CSharpSamplingExample
{
    #region private enumerations

    // input parameter type
    enum MyVariableType
    {
        // defined by a range [lowerbound;upperbound] where lowerbound and upperbound are real values
        Continuous = 0,
        // defined by a sorted list L={v_1;v_2;...;v_N} where L[i] is a real value for i=1..N, and L[i]<L[i+1] for i=1..N-1
        Continuouslist = 1,
        // defined by a list L={v_1;v_2;...;v_N} where L[i] is an integer value for i=1..N, and L[i]!=L[j] for i,j=1..N and i!=j
        Integerlist = 2
    }

    // objective type
    enum MyObjectiveType
    {
        Minimize = 0,
        Maximize = 1,
        SeekTarget = 2
    }

    // constraint type
    enum MyConstraintType
    {
        LessThan = 0,
        GreaterThan = 1,
        EqualTo = 2,
        Between = 3
    }

    #endregion private enumerations

    // CSharpSampling generates a set of random points
    //
    // Input Settings:
    // - Sampling Problem:
    //      - sampling domain (input variables)
    // - Sampling settings:
    //      - random generator seed 
    //      - number of sample points to generate
    //
    // Output Settings:
    // - sampling points
    public class CSharpSampling : ISamplingMethod
    {
        // the method's name should not countain empty space
        public const string MethodName = "CSharpSampling";

        #region private members

        // Common Services: provided by DesignXplorer
        // IServices is used to:
        // - submit one or several points to calculate
        // - publish log message
        // - publish progress
        IServices _services = null;

        // My Sampling Settings
        int _numberOfSamples = 0;
        // Sample points
        static List<IPoint> _samples = new List<IPoint>();
        static List<IPoint> _randomsamples = new List<IPoint>();

        // My Sampling problem
        // List of parameters id
        List<string> _parameters = new List<string>();

        // Dictionary of parameter types
        Dictionary<string, MyVariableType> _parametertype = new Dictionary<string, MyVariableType>();

        // Dictionary of parameter values
        // _parametersvalues[id] has 2 values (lower and upper bounds) when _parametertype[id] == Continuous
        // _parametersvalues[id] has a sorted list of real values when _parametertype[id] == Continuouslist
        // _parametersvalues[id] has a list of integer values when _parametertype[id] == Integerlist
        Dictionary<string, List<object>> _parametersvalues = new Dictionary<string, List<object>>();

        // True when there is one parameter relationship enabled at least
        private bool _hasParameterRelationship = false;

        // random generator seed 
        int _seedValue = 0;
        // random generator
        Random _random = null;
        // Max Min Distance
        private double _maxmindist = 0;
        // random point index
        int _ptIdx = 0;

        #endregion private members

        #region Constructors

        public CSharpSampling()
        {
        }

        #endregion Constructors

        #region ISamplingMethod

        // Services are initialized by DesignXplorer at the beginning of the Sampling run
        public IServices Services
        {
            set
            {
                _services = value;
            }
        }

        public object get_Setting(string propertyName)
        {
            object propertyValue;
            switch (propertyName)
            {
                case "NumberOfSamples":
                    propertyValue = _numberOfSamples;
                    break;
                case "SeedValue":
                    propertyValue = _seedValue;
                    break;
                case "MinDistance":
                    propertyValue = _maxmindist;
                    break;
                default:
                    propertyValue = null;
                    break;
            }

            return propertyValue;
        }

        public void put_Setting(string propertyName, object propertyValue)
        {
            switch (propertyName)
            {
                case "NumberOfSamples":
                    _numberOfSamples = (int)propertyValue;
                    break;
                case "SeedValue":
                    _seedValue = (int)propertyValue;
                    string msg = "Random Generator Initialized with seed value = " + _seedValue + "\n";
                    _services.PublishLogMessage(msg);
                    _random = new Random(_seedValue);
                    break;
            }
        }

        // definition of double variable
        // certain Sampling starts from an initial point (gradient based method for example)
        // initialvalue corresponds to the current value of the variable for the initial point
        // Ignore initialvalue if the Sampling does not start from initial point
        public void AddDoubleVariable(string id, double lowerbound, double upperbound, double initialvalue)
        {
            _parameters.Add(id);
            _parametertype.Add(id, MyVariableType.Continuous);
            List<object> listofvalues = new List<object>();
            listofvalues.Add(lowerbound);
            listofvalues.Add(upperbound);
            _parametersvalues.Add(id, listofvalues);
        }

        // definition of integer list variable
        // certain Sampling starts from an initial point (gradient based method for example)
        // initialvalue corresponds to the current value of the variable for the initial point
        // Ignore initialvalue if the Sampling does not start from initial point
        public void AddIntegerListVariable(string id, IListCpp values, int initialvalue)
        {
            _parameters.Add(id);
            _parametertype.Add(id, MyVariableType.Integerlist);
            List<object> listofvalues = new List<object>();
            int count = values.Count;
            for (int i = 0; i < count; i++ )
                listofvalues.Add(values[i]);
            _parametersvalues.Add(id, listofvalues);
        }

        // definition of double list variable
        // certain Sampling starts from an initial point (gradient based method for example)
        // initialvalue corresponds to the current value of the variable for the initial point
        // Ignore initialvalue if the Sampling does not start from initial point
        public void AddDoubleListVariable(string id, IListCpp values, double initialvalue)
        {
            _parameters.Add(id);
            _parametertype.Add(id, MyVariableType.Continuouslist);
            List<object> listofvalues = new List<object>();
            int count = values.Count;
            for (int i = 0; i < count; i++)
                listofvalues.Add(values[i]);
            _parametersvalues.Add(id, listofvalues);
        }

        // definition of output parameter
        // the problem can have several output parameters
        //
        // when any point is calculated, DesignXplorer set values on all output parameters in the sampling point
        // you can store all output values, 
        // or store only objectives and constraint values and ignore output parameters without objectives and constraints
        //
        public void AddOutput(string outputid)
        {
            // nothing
        }

        public void AddParameterRelationship(string variableid, string leftexpresssion, string rightexpression, enumParameterRelationshipType type)
        {
            _hasParameterRelationship = true;
        }

        // CanRun is called just before to run the Sampling, to check the definition of the sampling problem
        // if the Sampling can not be run, you can specify the reason in the returned message
        // this message will be displayed in the message view of the interface
        public bool CanRun(out string message)
        {
            // check of the number of variables
            if (_parameters.Count == 0)
            {
                message = "You need to specify one parameter at least.";
                return false;
            }

            // check the number of sample points to generate
            if (_numberOfSamples <= 0 || _numberOfSamples > 10000)
            {
                message = "Number of Samples must be positive and lower than or equal to 10000.";
                return false;
            }

            if (_hasParameterRelationship)
            {
                message = "Parameter Relationships not supported.";
                return false;
            }

            message = string.Empty;
            return true;
        }

        // Sampling
        public void Run()
        {
            IListCpp existingdps = _services.ExistingPoints;
            int nbexistingpts = (existingdps == null) ? 0 : existingdps.Count;

            _samples = new List<IPoint>();
            List<IPoint> allpts = new List<IPoint>();
            for (int ie = 0; ie < nbexistingpts; ie++)
            {
                IPoint pt = existingdps[ie] as IPoint;
                allpts.Add(pt);
            }
            string msg = "Number of Existing Points = " + nbexistingpts.ToString() + "\n";
            _services.PublishLogMessage(msg);

            // Publication of log message
            _services.PublishLogMessage("Generation of random points...\n");
            // generation of random points
            int iNumFeasiblePts = 0;
            for (int i = 0; i < _numberOfSamples*10 && iNumFeasiblePts < _numberOfSamples*2; i++)
            {
                IPoint point = GenerateRandomPoint();
                //if (_hasParameterRelationship)
                //{
                //    _services.CalculateParameterRelationshipsFromPoint(point);
                //    if (point.State == enumPointState.ePS_Invalid
                //        || point.State == enumPointState.ePS_Failed)
                //        continue;
                //}

                _randomsamples.Add(point);
                //_samples.Add(point);
                iNumFeasiblePts++;
            }

            int nbdps = Math.Min(iNumFeasiblePts, _numberOfSamples - nbexistingpts);
            for (int ip = 0; ip < nbdps; ip++)
            {
                if (allpts.Count == 0)
                {
                    IPoint pt = _randomsamples[0];
                    allpts.Add(pt);
                    _samples.Add(pt);
                    _randomsamples.RemoveAt(0);
                    continue;
                }

                double dmax2 = 0;
                int ibestpt = 0;
                for (int ir = 0; ir < _randomsamples.Count; ir++)
                {
                    IPoint randpt = _randomsamples[ir];
                    double dmin2 = Double.MaxValue;
                    for (int ia = 0; ia < allpts.Count && dmin2 > dmax2; ia++)
                    {
                        IPoint pt = allpts[ia];
                        double dist2 = GetDist2(randpt, pt);
                        if (dist2 < dmin2)
                            dmin2 = dist2;
                    }

                    if (dmin2 > dmax2)
                    {
                        dmax2 = dmin2;
                        ibestpt = ir;
                    }
                }

                IPoint bestpt = _randomsamples[ibestpt];
                allpts.Add(bestpt);
                _samples.Add(bestpt);
                _randomsamples.RemoveAt(ibestpt);
            }

            _maxmindist = MinDistance();

            if (iNumFeasiblePts == 0)
            {
                _services.PublishLogMessage("No feasible point!\n");
                return;
            }
        }

        public IListCpp Samples
        {
            get
            {
                if (_samples == null || _samples.Count == 0)
                    return null;

                IListCpp points = new VariantColl();
                int count = _samples.Count;
                for (int ic = 0; ic < count; ic++)
                    points.Add(_samples[ic]);
                return points;
            }
        }

        #endregion ISamplingMethod

        #region private methods

        // creates and generates a random point
        // only input variables values are defined
        private IPoint GenerateRandomPoint()
        {
            IPoint point = new DXOptimizationPoint();
            point.ID = "DP" + Convert.ToString(_ptIdx, System.Globalization.CultureInfo.InvariantCulture);
            _ptIdx++;

            if (_random == null)
            {
                string msg = "Random Generator Initialized with seed value = " + _seedValue + "\n";
                _services.PublishLogMessage(msg);
                _random = new Random(_seedValue);
            }

            int paramCount = _parameters.Count;
            for (int ip=0; ip<paramCount ; ip++)
            {
                string param = _parameters[ip];
                switch (_parametertype[param])
                {
                    case MyVariableType.Continuous:
                        {
                            double rand = _random.NextDouble();
                            double lb = (double) _parametersvalues[param][0];
                            double ub = (double) _parametersvalues[param][1];
                            double val = lb + rand*(ub - lb);
                            point.put_Value(param, val);
                        }
                        break;
                    case MyVariableType.Continuouslist:
                        {
                            int ub = (int) _parametersvalues[param].Count();
                            int idx = _random.Next(ub);

                            double val = (double) _parametersvalues[param][idx];
                            point.put_Value(param, val);
                        }
                        break;
                    case MyVariableType.Integerlist:
                        {
                            int ub = (int) _parametersvalues[param].Count();
                            int idx = _random.Next(ub);

                            int val = (int) _parametersvalues[param][idx];
                            point.put_Value(param, val);
                        }
                        break;

                }
            }
            return point;
        }

        private double MinDistance()
        {
            double dmin2 = double.MaxValue;
            double dist2;
            double dist;
            int nbsamples = _samples.Count;
            int nbparams = _parameters.Count;
            IPoint pti, ptj;
            for (int i = 1; i < nbsamples; i++)
            {
                pti = _samples[i];
                dmin2 = Double.MaxValue;
                for (int j = 0; j < i; j++)
                {
                    dist2 = 0;
                    ptj = _samples[j];
                    for (int ip = 0; ip < nbparams && dist2<dmin2; ip++)
                    {
                        string param = _parameters[ip];
                        dist = (double) pti.get_Value(param) - (double) ptj.get_Value(param);
                        dist2 += (dist*dist);
                    }
                    if (dist2 < dmin2)
                        dmin2 = dist2;
                }
            }

            return Math.Sqrt(dmin2);
        }

        private double GetDist2(IPoint pti, IPoint ptj)
        {
            double dist, dist2 = 0;
            int nbparams = _parameters.Count;

            for (int ip = 0; ip < nbparams; ip++)
            {
                string param = _parameters[ip];
                dist = (double)pti.get_Value(param) - (double)ptj.get_Value(param);
                dist2 += (dist * dist);
            }

            return dist2;
        }

        #endregion private methods
    }
}
